from __future__ import annotations

from typing import Any
from unittest.mock import Mock

import pytest
from jira import JIRA
from pytest_mock import MockerFixture

from firefighter.incidents.models.user import User
from firefighter.jira_app.models import JiraUser
from firefighter.raid.client import RaidJiraClient


class MockJiraClient(RaidJiraClient):
    def __init__(self, *args: Any, **kwargs: Any):
        pass

    @property
    def jira(self):
        return MockJiraAPI()


class MockJiraAPI(JIRA):
    def __init__(self, *args: Any, **kwargs: Any):
        pass

    def search_users(self, *args: Any, **kwargs: Any):
        return [Mock(raw={"accountId": "123", "emailAddress": "johndoe@example.com"})]

    def user(self, *args: Any, **kwargs: Any):
        return Mock(
            raw={
                "accountId": "123",
                "emailAddress": "johndoe@example.com",
                "displayName": "John Doe",
            }
        )


@pytest.fixture
def user() -> User:
    return User.objects.create(
        name="John Doe",
        email="johndoe@example.com",
        username="johndoe",
    )


@pytest.fixture
def not_user() -> User:
    return User.objects.create(
        name="John Does Not Exist",
        email="johndoeesnotexist@example.com",
        username="johndoeesnotexist",
    )


@pytest.fixture
def raid_client() -> MockJiraClient:
    return MockJiraClient()


@pytest.mark.django_db
def test_get_jira_user_from_user(
    raid_client: MockJiraClient, user: User, mocker: MockerFixture
):
    # Mock search_users
    raid_client.jira.search_users = Mock(
        return_value=[
            Mock(raw={"accountId": "123", "emailAddress": "johndoe@example.com"})
        ]
    )
    search_users = mocker.patch.object(raid_client.jira, "search_users")
    search_users.return_value = [
        Mock(raw={"accountId": "123", "emailAddress": "johndoe@example.com"})
    ]
    # Test when the user is valid and exists in Jira
    jira_user = raid_client.get_jira_user_from_user(user)
    assert jira_user.user.email == "johndoe@example.com"


@pytest.mark.django_db
def test_get_jira_user_from_user_in_db(
    raid_client: MockJiraClient, user: User, mocker: MockerFixture
):
    # Test when the user is valid and exists in Jira
    user.jira_user = JiraUser.objects.create(id="123", user=user)
    # Prepare no call jira.search_users(query=email_query)
    search_users = mocker.patch.object(raid_client.jira, "search_users")

    jira_user = raid_client.get_jira_user_from_user(user)
    assert jira_user.user.email == "johndoe@example.com"

    # Assert no call jira.search_users(query=email_query)
    search_users.assert_not_called()


@pytest.mark.django_db
def test_get_jira_user_from_jira_id(raid_client: MockJiraClient):
    # Test when the Jira ID is valid and exists in Jira
    jira_user = raid_client.get_jira_user_from_jira_id("123")
    assert jira_user.user.email == "johndoe@example.com"


@pytest.mark.django_db
def test_get_jira_user_from_jira_id_user_already_exists(
    raid_client: MockJiraClient, user: User
):
    # Test when the Jira ID is valid and exists in Jira
    jira_user = raid_client.get_jira_user_from_jira_id("123")
    assert jira_user.user.email == "johndoe@example.com"


@pytest.mark.django_db
def test_get_jira_user_from_jira_id_invalid_id(raid_client: MockJiraClient, user: User):
    with pytest.raises(ValueError, match="Jira account id is empty"):
        raid_client.get_jira_user_from_jira_id(None)  # type: ignore
    with pytest.raises(ValueError, match="Jira account id is empty"):
        raid_client.get_jira_user_from_jira_id("")



from __future__ import annotations

import json
import logging
from functools import cached_property
from typing import TYPE_CHECKING, Any

from django import apps
from django.contrib import admin
from django.contrib.admin import AdminSite, helpers
from django.contrib.admin.decorators import action
from django.contrib.admin.utils import model_ngettext
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.messages import constants
from django.template.response import TemplateResponse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext
from slack_sdk.errors import SlackApiError

from firefighter.incidents.models import (
    Component,
    Environment,
    Group,
    Incident,
    IncidentUpdate,
    Severity,
    User,
)
from firefighter.incidents.models.impact import (
    Impact,
    ImpactLevel,
    ImpactType,
    IncidentImpact,
)
from firefighter.incidents.models.incident_cost import IncidentCost
from firefighter.incidents.models.incident_cost_type import IncidentCostType
from firefighter.incidents.models.incident_membership import (
    IncidentMembership,
    IncidentRole,
)
from firefighter.incidents.models.incident_role_type import IncidentRoleType
from firefighter.incidents.models.metric_type import IncidentMetric, MetricType
from firefighter.incidents.models.milestone_type import MilestoneType
from firefighter.incidents.models.priority import Priority

if TYPE_CHECKING:
    from collections.abc import MutableSequence
    from decimal import Decimal

    from django.contrib.admin.options import (
        InlineModelAdmin,
        _ActionCallable,
        _FieldsetSpec,
    )
    from django.db.models.query import QuerySet
    from django.http.request import HttpRequest
    from django.utils.datastructures import _ListOrTuple

    from firefighter.slack.models.incident_channel import IncidentChannel

logger = logging.getLogger(__name__)

# Append inlines to these objects, from other modules
user_inlines: list[type[InlineModelAdmin[Any, User]]] = []
component_inlines: list[type[InlineModelAdmin[Any, Component]]] = []


class IncidentCostInline(admin.TabularInline[IncidentCost, Incident]):
    model = IncidentCost
    extra = 1
    fields = [
        "amount",
        "cost_type",
    ]


class IncidentImpactInline(admin.TabularInline[IncidentImpact, Incident]):
    model = IncidentImpact
    extra = 0
    raw_id_fields = ["impact"]


class IncidentMembershipInline(admin.StackedInline[IncidentMembership, Incident]):
    model = IncidentMembership
    extra: int = 0
    verbose_name = _("Incident Member")

    def has_change_permission(
        self,
        request: HttpRequest,  # noqa: ARG002
        obj: Any | None = None,  # noqa: ARG002
    ) -> bool:
        return False


incident_inlines: MutableSequence[type[InlineModelAdmin[Any, Incident]]] = []


@admin.register(Component)
class ComponentAdmin(admin.ModelAdmin[Component]):
    model = Component
    list_display = ["name", "group", "order", "private", "deploy_warning"]
    list_editable = ["order", "group", "private"]
    list_display_links = ["name"]
    list_filter = ["private", "deploy_warning", "group"]
    list_select_related = ["group"]
    ordering = ["group__order", "order"]
    search_fields = ["name", "group__name"]
    readonly_fields = [
        "created_at",
        "updated_at",
    ]
    inlines = component_inlines

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "name",
                    "description",
                    "group",
                    "order",
                    "private",
                    "deploy_warning",
                    "created_at",
                    "updated_at",
                )
            },
        ),
    )


class ComponentInline(admin.StackedInline[Component, Component]):
    model = Component
    fields = ["name", "order"]


class IncidentMetricInline(admin.TabularInline[IncidentMetric, Incident]):
    model = IncidentMetric
    extra = 0
    raw_id_fields = ["metric_type"]
    readonly_fields = ["duration", "metric_type", "incident"]
    can_delete = False

    @staticmethod
    def has_add_permission(*_: Any, **__: Any) -> bool:
        return False


@admin.register(Environment)
class EnvironmentAdmin(admin.ModelAdmin[Environment]):
    model = Environment
    list_display = [
        "value",
        "name",
        "description",
        "order",
    ]
    list_editable = [
        "order",
    ]
    list_display_links = ["value"]

    def get_readonly_fields(
        self,
        request: HttpRequest,  # noqa: ARG002
        obj: Environment | None = None,
    ) -> _ListOrTuple[str]:
        """Deny changing the value of an existing object."""
        if obj:  # editing an existing object
            return *self.readonly_fields, "value"
        return self.readonly_fields


@admin.register(Severity)
class SeverityAdmin(admin.ModelAdmin[Severity]):
    model = Severity
    list_display = [
        "name",
        "emoji",
        "description",
        "order",
        "enabled_create",
        "enabled_update",
    ]
    list_editable = [
        "order",
    ]
    list_display_links = ["name"]


@admin.register(Priority)
class PriorityAdmin(admin.ModelAdmin[Priority]):
    model = Priority
    list_display = [
        "name",
        "emoji",
        "description",
        "sla",
        "order",
        "enabled_create",
        "enabled_update",
    ]
    list_editable = [
        "order",
    ]
    list_display_links = ["name"]


class IncidentUpdateInline(admin.StackedInline[IncidentUpdate, Incident]):
    model = IncidentUpdate
    show_change_link = True
    fields = [
        "event_ts",
        "event_type",
        "_status",
        "priority",
        "message",
        "component",
        "created_by",
    ]
    extra = 0
    readonly_fields = [
        "priority",
        "message",
        "_status",
        "component",
        "created_by",
        "commander",
        "communication_lead",
    ]

    def get_queryset(self, request: HttpRequest) -> QuerySet[IncidentUpdate]:
        # TODO Optimize DB queries from IncidentUpdate admin
        return (
            super()
            .get_queryset(request)
            .select_related(
                "priority",
                "component",
                "incident",
                "incident__priority",
                "incident__component",
                "component__group",
                "created_by",
                "commander",
                "communication_lead",
            )
        )


@admin.register(Incident)
class IncidentAdmin(admin.ModelAdmin[Incident]):
    model = Incident

    @action(description=_("Compute metrics from milestones"))
    def compute_metrics(
        self, request: HttpRequest, queryset: QuerySet[Incident]
    ) -> None:
        for incident in queryset:
            incident.compute_metrics()
        self.message_user(
            request,
            ngettext(
                "Computed metrics for %d incident.",
                "Computed metrics for %d incidents.",
                len(queryset),
            )
            % len(queryset),
            constants.SUCCESS,
        )

    @action(description=_("Compute and purge metrics from milestones"))
    def compute_and_purge_metrics(
        self, request: HttpRequest, queryset: QuerySet[Incident]
    ) -> None:
        """Will compute metrics for selected incidents and delete metrics that can no longer be computed."""
        for incident in queryset:
            incident.compute_metrics(purge=True)
        self.message_user(
            request,
            ngettext(
                "Computed/purged metrics for %d incident.",
                "Computed/purged metrics for %d incidents.",
                len(queryset),
            )
            % len(queryset),
            constants.SUCCESS,
        )

    @action(
        description=_("Send message on conversation"),
    )
    def send_message(
        self, request: HttpRequest, queryset: QuerySet[Incident]
    ) -> TemplateResponse | None:
        """Action to send a message in selected channels.
        This action first displays a confirmation page to enter the message.
        Next, it sends the message on all selected objects and redirects back to the change list (other fn).
        """
        opts = self.model._meta  # noqa: SLF001

        # Get all the targeted conversations
        target_conversations: list[IncidentChannel] = [
            inc.conversation for inc in queryset.all() if inc.conversation
        ]

        deletable_objects = target_conversations

        # The user has already entered the messages.
        # Send message(s) and return None to display the change list view again.
        if request.POST.get("post"):
            self.process_action_send_message(
                request=request, target_conversations=target_conversations
            )
            return None

        objects_name = model_ngettext(queryset)

        title = _("Type your message")

        context = {
            **self.admin_site.each_context(request),
            "title": title,
            "objects_name": str(objects_name),
            "deletable_objects": [deletable_objects],
            "queryset": queryset,
            "opts": opts,
            "action_checkbox_name": helpers.ACTION_CHECKBOX_NAME,
            "media": self.media,
        }

        # Display the message edit page
        return TemplateResponse(
            request, "admin/send_message_conversation.html", context
        )

    def process_action_send_message(
        self, request: HttpRequest, target_conversations: list[IncidentChannel]
    ) -> None:
        # TODO Add and check permissions
        # if perms_needed:
        #     raise PermissionDenied

        message_text = request.POST.get("text")
        message_blocks_raw = request.POST.get("blocks")
        if message_text == "":
            message_text = None
        if not message_blocks_raw:
            message_blocks = None
        else:
            message_blocks = json.loads(message_blocks_raw)
            if isinstance(message_blocks, list) or all(
                isinstance(block, dict) for block in message_blocks
            ):
                raise ValueError("Blocks must be a list of JSON dicts")

        if message_blocks is None and message_text is None:
            raise ValueError("You need texts or blocks to send a message!")

        success = []
        errors = []

        for conversation in target_conversations:
            try:
                conversation.send_message(text=message_text, blocks=message_blocks)
                success.append(((conversation.name, conversation.incident_id), True))
            except SlackApiError as e:
                errors.append(((conversation.name, conversation.incident_id), False, e))

        if len(success) > 0:
            # XXX Test still working!
            self.message_user(
                request,
                ngettext(
                    _("Sent message on %d conversation (%s).")
                    % {", ".join(f"#{key[0][0]}, #{key[0][1]}" for key in success)},
                    f"Sent messages on %d conversations ({', '.join(f'#{key[0][0]}, #{key[0][1]}' for key in success)}).",
                    len(success),
                )
                % len(success),
                constants.SUCCESS,
            )
        if len(errors) > 0:
            self.message_user(
                request,
                format_html(
                    "Error sending message: <br/>{}",
                    "<br/>".join(
                        f"#{key[0][0]}, #{key[0][1]}: {key[2]}" for key in errors
                    ),
                ),
                constants.ERROR,
            )

        # Return None to display the change list page again.

    date_hierarchy = "created_at"
    autocomplete_fields = ["created_by", "component"]
    actions: list[_ActionCallable[Any, Incident]] = [
        compute_metrics,
        compute_and_purge_metrics,
        send_message,
    ]
    list_display = [
        "id",
        "title",
        "short_description",
        "_status",
        "priority",
        "component",
        "environment",
        "created_at",
        "updated_at",
    ]

    list_display_links = ["id", "title"]
    list_filter = ("_status", "priority", "component", "environment")
    readonly_fields = (
        "created_at",
        "updated_at",
        "can_be_closed",
        "ask_for_milestones",
        "missing_milestones",
        "total_cost",
    )

    list_select_related = (
        "priority",
        "component__group",
        "environment",
    )
    list_max_show_all = 1000

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "title",
                    "description",
                    "_status",
                    "ignore",
                    "private",
                    "created_at",
                    "updated_at",
                    "tags",
                )
            },
        ),
        (_("Relations"), {"fields": ("priority", "component", "environment")}),
        (
            _("User Roles"),
            {
                "fields": ("created_by",),
            },
        ),
        (
            _("Properties"),
            {
                "classes": ("collapse",),
                "fields": (
                    "can_be_closed",
                    "ask_for_milestones",
                    "missing_milestones",
                    "severity",
                ),
            },
        ),
    )

    incident_inlines.append(IncidentUpdateInline)
    incident_inlines.append(IncidentMembershipInline)
    incident_inlines.append(IncidentCostInline)
    incident_inlines.append(IncidentImpactInline)
    incident_inlines.append(IncidentMetricInline)
    inlines = incident_inlines  # type: ignore[assignment]
    search_fields: list[str] = [
        "title",
        "description",
    ]  # We need to define it even if we redefine get_search_results so that the input is shown in the admin

    @staticmethod
    def total_cost(obj: Incident) -> int | float | Decimal:
        return obj.total_cost

    @staticmethod
    def get_search_results(
        request: HttpRequest,  # noqa: ARG004
        queryset: QuerySet[Incident],
        search_term: str,
    ) -> tuple[QuerySet[Incident], bool]:
        return Incident.objects.search(queryset=queryset, search_term=search_term)

    def get_queryset(self, request: HttpRequest) -> QuerySet[Incident]:
        return (
            super()
            .get_queryset(request)
            .select_related(*self._get_select_related)
            .prefetch_related()
        )

    @cached_property
    def _get_select_related(self) -> list[str]:
        select_related = [
            "priority",
            "component__group",
            "environment",
            "conversation",
            "component",
            "created_by",
        ]
        if apps.apps.is_installed("firefighter.confluence"):
            select_related.append("postmortem_for")
        return select_related


@admin.register(IncidentUpdate)
class IncidentUpdateAdmin(admin.ModelAdmin[IncidentUpdate]):
    model = IncidentUpdate
    list_display = [
        "incident_id",
        "_status",
        "priority",
        "message",
        "component",
        "created_by",
    ]
    list_display_links = ["_status"]
    list_filter = ("_status", "priority", "component", "event_type")
    readonly_fields = ("created_at", "updated_at", "incident")
    list_select_related = ("priority", "component", "incident", "created_by")
    search_fields = ["description", "message"]


@admin.register(Group)
class GroupAdmin(admin.ModelAdmin[Group]):
    model = Group
    list_display = ["name", "order"]
    list_editable = [
        "order",
    ]
    ordering = ["order"]
    list_display_links = ["name"]
    inlines = [ComponentInline]
    search_fields = ["name"]


@admin.register(User)
class UserAdmin(BaseUserAdmin):  # type: ignore[type-arg]
    model = User
    list_max_show_all = 500
    inlines = user_inlines
    readonly_fields = [
        "commander_count",
        "communication_lead_count",
        "incidents_opened_count",
    ]
    list_display_links = ["email"]
    ordering = ["first_name", "last_name"]
    list_display = ("username", "email", "full_name", "is_staff", "is_active", "bot")
    list_filter = ("is_staff", "is_superuser", "is_active", "groups", "bot")

    def __init__(self, model: type[User], admin_site: AdminSite) -> None:
        super().__init__(model, admin_site)
        self.fieldsets[2][1]["fields"] = ("bot", *self.fieldsets[2][1]["fields"])  # type: ignore
        self.fieldsets[1][1]["fields"] = (*self.fieldsets[1][1]["fields"], "avatar")  # type: ignore

    @staticmethod
    def commander_count(obj: User) -> int:
        return obj.roles_set.filter(role_type__slug="commander").count()

    @staticmethod
    def communication_lead_count(obj: User) -> int:
        return obj.roles_set.filter(role_type__slug="communication_lead").count()

    @staticmethod
    def incidents_opened_count(obj: User) -> int:
        return obj.incidents_created_by.count()

    def get_fieldsets(
        self,
        request: HttpRequest,
        obj: User | None = None,
    ) -> _FieldsetSpec:
        fieldsets = list(super().get_fieldsets(request, obj))
        fieldsets.append(
            (
                _("User statistics"),
                {
                    "fields": (
                        "commander_count",
                        "communication_lead_count",
                        "incidents_opened_count",
                    )
                },
            )
        )
        return fieldsets


@admin.register(IncidentMembership)
class IncidentMembershipAdmin(admin.ModelAdmin[IncidentMembership]):
    model = IncidentMembership
    fields = ("user", "incident")
    list_display = ("user", "incident")


class ImpactLevelInline(admin.TabularInline[ImpactLevel, Incident]):
    model = ImpactLevel
    extra = 0


@admin.register(IncidentCost)
class IncidentCostAdmin(admin.ModelAdmin[IncidentCost]):
    model = IncidentCost
    fields = ("incident", "amount", "cost_type", "details")
    list_display = ("incident", "amount", "cost_type")
    search_fields = ["incident__id", "incident__title", "amount", "cost_type"]
    autocomplete_fields = ["incident"]


@admin.register(IncidentCostType)
class IncidentCostTypeAdmin(admin.ModelAdmin[IncidentCostType]):
    model = IncidentCostType
    list_display = ("name", "description")


@admin.register(Impact)
class ImpactAdmin(admin.ModelAdmin[Impact]):
    model = Impact


@admin.register(ImpactType)
class ImpactTypeAdmin(admin.ModelAdmin[ImpactType]):
    model = ImpactType
    prepopulated_fields = {"value": ["name"]}

    list_display = ("name", "value", "help_text")
    search_fields = ["name", "value", "help_text"]

    inlines = [ImpactLevelInline]


@admin.register(ImpactLevel)
class ImpactLevelAdmin(admin.ModelAdmin[ImpactLevel]):
    model = ImpactLevel
    list_display = (
        "value",
        "name",
        "impact_type",
    )
    search_fields = [
        "name",
        "impact_type",
    ]
    list_filter = ("impact_type", "value")


@admin.register(MilestoneType)
class MilestoneTypeAdmin(admin.ModelAdmin[MilestoneType]):
    model = MilestoneType
    list_display = (
        "name",
        "user_editable",
        "required",
        "asked_for",
    )

    search_fields = ["name", "description", "summary"]
    prepopulated_fields = {"event_type": ["name"], "summary": ["name"]}


@admin.register(MetricType)
class MetricTypeAdmin(admin.ModelAdmin[MetricType]):
    model = MetricType
    list_display = (
        "name",
        "code",
        "milestone_lhs",
        "milestone_rhs",
    )

    search_fields = ["name", "code"]
    prepopulated_fields = {"type": ["name"]}


@admin.register(IncidentMetric)
class IncidentMetricAdmin(admin.ModelAdmin[IncidentMetric]):
    model = IncidentMetric
    list_display = (
        "metric_type",
        "duration",
        "incident",
    )
    search_fields = ["incident__title", "metric_type__title", "value"]
    list_filter = ("metric_type",)


@admin.register(IncidentRoleType)
class IncidentRoleTypeAdmin(admin.ModelAdmin[IncidentRoleType]):
    """IncidentRoleTypes are fixed at the moment (only commander and communication lead)."""

    list_display = ["name", "emoji", "required", "order"]
    readonly_fields = ["id"]
    ordering = ["order"]
    fieldsets = [
        (
            _("⚙️ General Information"),
            {"fields": ["id", "slug", "name", "emoji", "required", "order"]},
        ),
        (
            _("📝 Short Descriptions"),
            {"fields": ["summary", "summary_first_person", "description", "aka"]},
        ),
    ]
    search_fields = ["name", "slug", "emoji", "summary", "description"]
    list_filter = ["required"]
    prepopulated_fields = {"slug": ["name"]}

    @staticmethod
    def has_delete_permission(
        _request: HttpRequest, _obj: IncidentRoleType | None = None
    ) -> bool:
        return False


@admin.register(IncidentRole)
class IncidentRoleAdmin(admin.ModelAdmin[IncidentRole]):
    model = IncidentRole
    list_display = (
        "incident",
        "user",
        "role_type",
    )
    search_fields = ["incident__title", "user__email", "role_type__name"]
    list_filter = ("role_type",)
    autocomplete_fields = ["incident", "user", "role_type"]
    readonly_fields = ["incident", "user", "role_type"]



from __future__ import annotations

import logging
from functools import cache
from typing import Final

COLORS: dict[str, str] = {
    "Forest Green": "#086e3f",
    "Emerald Green": "#58ad72",
    "Lint": "#4bd676",
    "Pale Lint": "#75e097",
    "Fuchsia": "#fb4570",
    "Hot Pink": "#fb6b90",
    "Pink": "#fb8da0",
    "Pink White": "#efebe0",
    "Red": "#ed4040",
    "Ivory": "#f1ece4",
    "Nude": "#c3b090",
    "Sand Dollar": "#de943a",
    "Tan": "#92794f",
    "Blue Gray": "#8da7c4",
    "Sky": "#ace1fc",
    "Stone Blue": "#8da7c4",
    "White Blue": "#e5ddfc",
}

PALETTES = {
    logging.DEBUG: ["Blue Gray", "Sky", "Stone Blue", "White Blue"],
    logging.INFO: ["Forest Green", "Lint", "Emerald Green", "Pale Lint"],
    logging.WARNING: ["Nude", "Tan", "Nude", "Sand Dollar"],
    logging.ERROR: ["Hot Pink", "Fuchsia", "Pink", "Red"],
    logging.CRITICAL: ["Hot Pink", "Fuchsia", "Pink", "Red"],
}
# pylint: disable=consider-using-f-string

_ANSI_END: Final[str] = "\x1b[0m"


def _html_to_triple_dec(html_code: str) -> list[int]:
    return [int(x, 16) for x in (html_code[1:3], html_code[3:5], html_code[5:8])]


@cache
def _ansi_fg(name: str) -> str:
    html_code = COLORS[name]
    return "\x1b[38;2;{};{};{}m".format(*_html_to_triple_dec(html_code))


class PrettyFormatter(logging.Formatter):
    """A custom logging formatter that formats log records with colors and a pretty output, for local development.

    Opiniated about what is pretty.

    Attributes:
        _style (logging._STYLES): The logging style used for formatting.
        _fmt (str): The logging format string used for formatting.
    """

    def _set_format(self, fmt: str, style: str = "%") -> None:
        self._style = logging._STYLES[style][0](fmt)  # type: ignore[operator] # noqa: SLF001 # Mysticism 🤔
        self._fmt = self._style._fmt  # noqa: SLF001

    def format(self, record: logging.LogRecord) -> str:
        # Reference attributes: https://docs.python.org/3/library/logging.html#logrecord-attributes
        palette = PALETTES.get(record.levelno, PALETTES[logging.DEBUG])
        self._set_format(
            f"{_ansi_fg(palette[0])}%(asctime)s "
            f"{_ansi_fg(palette[1])}%(levelname)-8s "
            f"{_ansi_fg(palette[2])}%(name)s:%(lineno)d "
            f"{_ansi_fg(palette[3])}%(message)s"
            f"{_ANSI_END}"
        )
        return super().format(record)




import logging

from django.db import migrations

logger = logging.getLogger(__name__)


def get_component_mappings() -> list:
    """
    Returns a list of tuples for updating existing component names and their attributes.

    Each tuple contains:
        - old_name (str): The current name of the component.
        - new_name (str): The new name to assign to the component.
        - slack_channel (str): The associated Slack channel for the component.
        - group_name (str): The name of the group to which the component belongs.

    Returns:
        list: A list of tuples, each representing the details for a component update.
    """
    return [
        # Marketplace
        ("Product Discovery", "Navigation & Product discovery", "impact-nav-product-discovery", "Marketplace"),
        ("User & Purchase", "Cart & funnel", "impact-cart-funnel", "Marketplace"),
        ("Customer Management", "Customer login & signup", "impact-customer-login-signup", "Marketplace"),
        ("Performance", "Web performance", "impact-web-performance", "Marketplace"),
        # Catalog
        ("Product Information", "Product Management", "impact-product-management", "Catalog"),
        ("Taxonomy", "Product Structure", "impact-product-structure", "Catalog"),
        ("Publication on website", "Catalog Exposition", "impact-catalog-exposition", "Catalog"),
        # Payment Operations
        ("Payment", "Payment", "impact-payment", "Payment Operations"),
        # Operations
        ("ManoFulfillment OPS", "MM Fulfillment", "impact-mm-fulfillment", "Operations"),
        ("Helpcenter", "Helpcenter after sales", "impact-helpcenter-after-sales", "Operations"),
        # Finance
        ("Finance Operations", "Controlling", "impact-controlling", "Finance"),
        # Platform
        ("Spinak", "Spinak", "impact-spinak", "Platform"),
        ("CDN", "CDN", "impact-cdn", "Platform"),
        ("Gitlab", "Gitlab", "impact-gitlab", "Platform"),
        # Data
        ("data-platform", "Data Ingestion", "impact-data-ingestion", "Data"),
        ("data-specialist-offer", "Data Warehouse", "impact-data-warehouse", "Data"),
        ("data-wbr", "Data Analytics", "impact-data-analytics", "Data"),
        # Security
        ("Security Misc", "Bot management & rate limiting & WAF", "impact-bot-management-rate-limiting-waf", "Security"),
        ("Attack", "Data leak", "impact-data-leak", "Security"),
        ("System Compromise", "Exploited vulnerability", "impact-exploited-vulnerability", "Security"),
        ("Personal Data Breach", "Stolen account(s) or IT materials", "impact-stolen-accounts-it-materials", "Security"),
    ]


def update_component_names(apps, schema_editor):
    Component = apps.get_model("incidents", "Component")
    Group = apps.get_model("incidents", "Group")
    component_mappings = get_component_mappings()

    updated_count = 0

    for old_name, new_name, _slack_channel, group_name in component_mappings:
        try:
            component = Component.objects.get(name=old_name)
            logger.info(f"Updating: '{old_name}' to '{new_name}'")
            component.name = new_name

            group_instance = Group.objects.get(name=group_name)
            component.group = group_instance
            component.save()
            updated_count += 1
        except Exception:
            logger.exception(f"Component '{old_name}' does not exist, cannot proceed with updates.")


def revert_component_names(apps, schema_editor):
    Component = apps.get_model("incidents", "Component")
    reverse_mappings = {new_name: old_name for old_name, new_name, _, _ in get_component_mappings()}

    updated_count = 0

    for new_name, old_name in reverse_mappings.items():

        try:
            component = Component.objects.get(name=new_name)
            logger.info(f"Restoring '{new_name}' back to '{old_name}'")
            component.name = old_name
            component.save()
            updated_count += 1
        except Exception:
            logger.exception(f"Component '{new_name}' does not exist, skipping restoration.")


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0009_update_sla"),
    ]

    operations = [
        migrations.RunPython(update_component_names, revert_component_names),
    ]



from __future__ import annotations

import logging

from celery import shared_task
from django.db import transaction

from firefighter.pagerduty.models import PagerDutyService
from firefighter.pagerduty.service import pagerduty_service

logger = logging.getLogger(__name__)


@shared_task(name="pagerduty.fetch_services")
@transaction.atomic
def fetch_services() -> None:
    """Celery task to fetch PagerDuty services and save them in the database."""
    fetched_services_key = []
    for service in pagerduty_service.client.session.iter_all("services"):
        fetched_services_key.append(service["id"])
        PagerDutyService.objects.update_or_create(
            pagerduty_id=service["id"],
            defaults={
                "name": service["name"][:128],
                "status": service["status"][:128],
                "summary": service["summary"][:256],
                "web_url": service["html_url"][:256],
                "api_url": service["self"][:256],
            },
        )

    # Check that we don't have stale services
    if len(fetched_services_key) != PagerDutyService.objects.count():
        logger.warning("Stale PagerDuty Services found in DB. Manual action needed.")
        # XXX Delete stale services if we are confident enough



from __future__ import annotations

import logging
from itertools import groupby
from typing import TYPE_CHECKING, Any

from django import forms
from django.forms import ValidationError
from django_filters import Filter

from firefighter.incidents.views.date_filter import (
    get_date_range_from_special_date,
    get_range_look_args,
)

if TYPE_CHECKING:
    from datetime import date, datetime

    from django.db.models.query import QuerySet

logger = logging.getLogger(__name__)


class TextDateRangeField(forms.Field):
    def to_python(
        self, value: str | None
    ) -> tuple[date | None, date | None, str | None, str | None] | None:
        """Validate date range. Field is optional."""
        if not value or value.strip() == "":
            return None
        date_range = get_date_range_from_special_date(value)
        if date_range[0] is None and date_range[1] is None:
            err_msg = f"Invalid date range: {value}"
            raise ValidationError(err_msg)
        if (date_range[0] is not None and date_range[1] is not None) and date_range[
            0
        ] > date_range[1]:
            err_msg = f"Start date must be before end date ({date_range[0].strftime('%Y-%m-%d %H:%M:%S')} > {date_range[1].strftime('%Y-%m-%d %H:%M:%S')})."
            raise ValidationError(err_msg)

        return date_range


class FFDateRangeSingleFilter(Filter):
    """TODO Fix typings, implement tests, move in proper location."""

    field_class = TextDateRangeField

    def filter(self, qs: QuerySet[Any], value: Any) -> QuerySet[Any]:
        if not value:
            return qs

        qs = qs.filter(**self.get_lookup_kwarg(value))
        return qs.distinct() if self.distinct else qs

    def get_lookup_kwarg(
        self, value: tuple[datetime | None, datetime | None, Any, Any]
    ) -> dict[str, datetime]:
        field_name = self.field_name
        gte, lte, _, _ = value

        return get_range_look_args(gte, lte, field_name=field_name)


class FFDateRangeSingleFilterEmpty(FFDateRangeSingleFilter):
    def filter(self, qs: QuerySet[Any], value: Any) -> QuerySet[Any]:
        if not value:
            return qs

        return qs.distinct() if self.distinct else qs


class CustomCheckboxSelectMultiple(forms.widgets.CheckboxSelectMultiple):
    """Custom template."""

    template_name = "incidents/widgets/grouped_checkbox_nested.html"


class GroupedCheckboxSelectMultiple(CustomCheckboxSelectMultiple):
    """Widget to group checkboxes in a select multiple.
    TODO Make this generic!
    """

    def optgroups(
        self, name: str, value: list[Any], attrs: dict[str, Any] | None = None
    ) -> list[tuple[str | None, list[dict[str, Any]], int | None]]:
        """Return a list of optgroups for this widget."""
        self.choices = [
            (key, list(group))
            for key, group in groupby(self.choices, key=lambda x: x[0].instance.group)  # type: ignore[union-attr]
        ]
        return super().optgroups(name, value, attrs)



"""This file contains all the settings used in production environments.

This file is required and if `ENV=dev` these values are not used.
"""

from __future__ import annotations

from pathlib import Path

from firefighter.firefighter.settings.settings_utils import config

# Production flags:
# https://docs.djangoproject.com/en/4.2/howto/deployment/

DEBUG = config("DEBUG", default=False, cast=bool)

# Staticfiles
# https://docs.djangoproject.com/en/4.2/ref/contrib/staticfiles/

# This is a hack to allow a special flag to be used with `--dry-run`
# to test things locally.
_COLLECTSTATIC_DRYRUN = config(
    "DJANGO_COLLECTSTATIC_DRYRUN",
    cast=bool,
    default=False,
)
# Adding STATIC_ROOT to collect static files via 'collectstatic':
STATIC_ROOT = Path(
    ".static"
    if _COLLECTSTATIC_DRYRUN
    else config("STATIC_ROOT", default="/var/app/django/.static")
)

STORAGES = {
    "staticfiles": {
        "BACKEND": "django.contrib.staticfiles.storage.ManifestStaticFilesStorage",
    },
}


# Media files
# https://docs.djangoproject.com/en/4.2/topics/files/

MEDIA_ROOT = "/var/app/django/media"


# Security
# https://docs.djangoproject.com/en/4.2/topics/security/

SECURE_HSTS_SECONDS = 31536000  # the same as Caddy has
SECURE_HSTS_INCLUDE_SUBDOMAINS = True
SECURE_HSTS_PRELOAD = True

SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")
SECURE_SSL_REDIRECT = config("SECURE_SSL_REDIRECT", default=True, cast=bool)
SECURE_REDIRECT_EXEMPT = [
    # This is required for healthcheck to work:
    r"^api/v2/firefighter/monitoring/healthcheck$",
]

SESSION_COOKIE_SECURE = True
CSRF_COOKIE_SECURE = True
